'''
binfull.py
A module containing classes and functions for the binfull algorithm.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
from __future__ import division
import math
import numpy as np
import bisect
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from utilities.sampleFunctions import MyRandom


# binfull gets its own randomGenerator
randomGenerator = MyRandom()
randomGenerator.setSeed('default_two')


###### CDF GENERATION UTILITIES ######
def forwardSum(arg):
  """ forwardSum(arg):
      return np.array([arg[0]]
                    + [arg[i] + arg[i-1] for i in xrange(1, len(arg))]) """
  return np.array([arg[0]] + [arg[i+1] + arg[i] for i in xrange(len(arg)-1)])

def total(arg):
  """ total(arg):
      return np.cumsum(arg) """
  return np.cumsum(arg)

def A(arg, bins):
  """ A(arg, bins):
      if 'bins' specifies the bin edges as a numpy array:
        return 0.5 * total(numpy.diff(bins) * forwardSum(arg))
      else:
        return 0.5 * total(bins * forwardSum(arg)) """
  if type(bins) == np.ndarray:
    return 0.5 * total(np.diff(bins) * forwardSum(arg))
  return 0.5 * total(bins * forwardSum(arg))


###### SMOOTHERS FOR BINFULL ALGORITHM ######
class gaussianSmoother(object):
  """ Smoothing functor for use in dataToBinfull.
      Simple Gaussian random deviate smoothing function. """
  def __init__(self, sigma):
    """ Constructor. Input the value to be used for the Gaussian sigma. """
    self.sigma = sigma
  def __getitem__(self, index):
    return [0., 100.]
  def __call__(self, **kwargs):
    if self.sigma <= 0:
      return 0
    return randomGenerator.normal(0, self.sigma)
  def __str__(self):
    return 'Gaussian'

class linearExpandingGaussianSmoother(object):
  """ Smoothing functor for use in dataToBinfull.
      Simple Gaussian random deviate smoothing function. However,
      the value used for the sigma of that Gaussian starts from zero,
      but grows each time the same data value is smoothed. """
  def __init__(self, sigma):
    """ Constructor. Input the value to be used for the Gaussian sigma. """
    self.sigma = sigma
    self.counters = {}
  def __getitem__(self, index):
    return [0., 0.5]
  def __call__(self, **kwargs):
    if self.sigma <= 0:
      return 0
    j = kwargs['inJ']
    self.counters[j] = self.counters.setdefault(j, 0) + 1
    return randomGenerator.normal(0, self.sigma * self.counters[j])
  def __str__(self):
    return 'LinearExpanding'

class sqrtExpandingGaussianSmoother(object):
  """ Smoothing functor for use in dataToBinfull.
      Simple Gaussian random deviate smoothing function. However,
      the value used for the sigma of that Gaussian starts from zero,
      but grows each time the same data value is smoothed. """
  def __init__(self, sigma):
    """ Constructor. Input the value to be used for the Gaussian sigma. """
    self.sigma = sigma
    self.counters = {}
  def __getitem__(self, index):
    return [0., 10.]
  def __call__(self, **kwargs):
    if self.sigma <= 0:
      return 0
    j = kwargs['inJ']
    n = kwargs['inN']
    self.counters[j] = self.counters.setdefault(j, 0) + 1
    return randomGenerator.normal(0, self.sigma * math.sqrt(self.counters[j]))
  def __str__(self):
    return 'SqrtExpanding'

smootherTuple = (gaussianSmoother, linearExpandingGaussianSmoother,
                 sqrtExpandingGaussianSmoother)


###### BINFULL MONTE CARLO ALGORITHM ######
def dataToBinfull(data, nbinfull=10000, smoother=None):
  """ dataToBinfull(data, nbinfull=10000, smoother=None):
      Takes a list of data values to form a cdf and binfull histogram.
      'smoother' -- see smoothing functors in smootherTuple.
      Generates as huge numpy arrays:
        cdf - values from the cumulative distribution of the data.
        binfull - nbinfull values generated from the cdf.
      (See the BinfullHistogram class if you want a more lightweight output.)
      Returns (cdf, binfull) """
  data.sort()
  cdf = np.arange(1., len(data) + 1, 1.)
  # bisect_left(cdf, p) locates the left-most entry in cdf which is >= p
  indices = [bisect.bisect_left(cdf, p)
             for p in (randomGenerator.rand(nbinfull) * cdf[-1])]
  if not isinstance(smoother, smootherTuple):
    binfull = np.array([data[j] for j in indices])
  else:
    binfull = np.array([smoother(inJ=j, inN=nbinfull, inCdf=cdf, inData=data)
                        + data[j] for j in indices])
  return (cdf, binfull)

class binfullHistogram(object):
  """ The results of a binfull algorithm could be stored as a numpy array.
      (For an example, see the dataToBinfull function.)
      However, due to the huge number of data points generated, it becomes a
      memory issue. This container is more memory lightweight. """
  def __init__(self, binEdges=None, data=None, nbinfull=10000, smoother=None):
    """ binfullHistogram(binEdges, data, nbinfull=10000, smoother=None)
        [constructor]
        Uses the binfull algorithm to construct the binfullHistogram. 
        'binEdges' -- a numpy array of the locations of the bin edges.
        See 'dataToBinfull' for more information. """
    if binEdges != None and data != None:
      self.binEdges = binEdges
      self.histo = np.zeros(len(binEdges)-1)  # array of bin contents
      data.sort()
      cdf = np.arange(1., len(data) + 1, 1.)
      # NOTE: no longer uses (as much) the power of the numpy arrays... Slow??
      # bisect_left(cdf, p) locates the left-most entry in cdf which is >= p
      for p in (randomGenerator.rand(nbinfull) * cdf[-1]):
        j = bisect.bisect_left(cdf, p)
        # bisect_right(self.binEdges, x)-1 locates the bin to fill.
        if not isinstance(smoother, smootherTuple):
          fillIndex = bisect.bisect_right(self.binEdges, data[j])-1
        else:
          fillIndex = bisect.bisect_right(self.binEdges, data[j]
              + smoother(inJ=j, inN=nbinfull, inCdf=cdf, inData=data))-1
        if fillIndex >= 0 and fillIndex < len(self.histo):
          self.histo[fillIndex] += 1

  def __len__(self):
    return len(self.histo)
  def integral(self):
    """ Returns the total integration over the entire binfull pdf. """
    return A(self.histo, self.binEdges)[-1]
  def fillPlot(self, **kwargs):
    """ Plots the result, filling under the 'histogram'. 
        Passes **kwargs on to the matplotlib.pyplot.fill_between function.
        Returns a proxy matplotlib.patches.Rectangle for legend control. """
    x = np.sort(np.concatenate((self.binEdges, self.binEdges)))
    y = np.array([[value, value] for value in self.histo / self.integral()])
    y = np.concatenate(([1e-6], y.flatten(), [1e-6]))
    plt.fill_between(x, y, 1e-6, **kwargs)
    return Rectangle((0, 0), 1, 1, fill=True, **kwargs)
  def plot(self, fmtString='ro', **kwargs):
    """ Plots the result, as probability vs bin center.
        First argument is an optional format string. Also passes along
        **kwargs to the matplotlib.pyplot.plot function.
        Returns the plot for legend control. """
    binCenters = np.diff(self.binEdges)/2 + self.binEdges[:-1]
    p, = plt.plot(binCenters, self.histo / self.integral(), fmtString, **kwargs)
    return p

def binfullFactory(binEdges, histo):
  """ binfullFactory(binEdges, histo):
      Create a binfullHistogram instance from loaded binfull data. """
  result = binfullHistogram()
  result.binEdges = binEdges
  result.histo = histo
  return result
