'''
binless.py
A module containing functions for the binless algorithm.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
import numpy as np
from scipy.sparse import lil_matrix, spdiags


###### DATA SET UP UTILITIES ######
def uniqueSort(x):
  """ uniqueSort(x):
      return np.sort(np.unique(np.array(x))) """
  return np.sort(np.unique(np.array(x)))

def dataCdf(x):
  """ dataCdf(x):
      return np.arange(len(x)) + 1. """
  return np.arange(len(x)) + 1.


###### NUMERICAL CALCULUS UTILITIES ######
def makeForwardDiff(n, boundaryCondition='periodic'):
  """ makeForwardDiff(n):
      Creates a 'forward difference' matrix which is n by n, which is useful
      for constructing a matrix derivative operator.
      If n is a numpy.ndarray, the matrix is instead len(n) by len(n). """
  if type(n) == np.ndarray:
    n = len(n)
  if boundaryCondition == 'periodic':
    # initialize as a linked-list sparse matrix
    forwardDiff = lil_matrix((n, n))
    # add 1's to the diagonal
    forwardDiff.setdiag([1.0] * n)
    # add -1's to the subdiagonal
    forwardDiff.setdiag([-1.0] * (n - 1), k=-1)
    # add -1 to the upper-right corner
    forwardDiff[0, n-1] = -1.0
    # convert to Compressed Sparse Row format
    forwardDiff = forwardDiff.tocsr()
  # TODO: implement derivative matrices with more boundary conditions!!
  else:
    raise ValueError('no prescription for boundary condition: ' + boundaryCondition)
  return forwardDiff

def makeDerivativeArrays(x):
  """ makeDerivativeArrays(x):
      Returns (dx, forwardDiff, delForward), where
        dx is a 'delta x' numpy.ndarray,
        forwardDiff is a 'forward difference' matrix [see makeForwardDiff],
        and delForward is a first derivative matrix operator.
      x should be a sorted numpy.ndarray of data values. """
  dx = np.concatenate(( [x[0]], np.diff(x) ))
  n = len(x)
  forwardDiff = makeForwardDiff(n)
  delForward = spdiags(dx**(-1.0), [0], n, n) * forwardDiff
  return (dx, forwardDiff, delForward)

def forwardSum(arg):
  """ forwardSum(arg):
      return np.array([arg[0]]
                    + [arg[i] + arg[i-1] for i in xrange(1, len(arg))]) """
  return np.array([arg[0]] + [arg[i] + arg[i-1] for i in xrange(1, len(arg))])

def total(arg):
  """ total(arg):
      return np.cumsum(arg) """
  return np.cumsum(arg)

def totalT(arg):
  """ totalT(arg):
      # flip array except first element. add 0 as first element.
      # cumulative sum, then flip again.
      return np.cumsum([0] + list(arg[:0:-1]))[::-1] """
  return np.cumsum([0] + list(arg[:0:-1]))[::-1]

def A(arg, dx):
  """ A(arg, dx):
      return 0.5 * total(dx * forwardSum(arg)) """
  return 0.5 * total(dx * forwardSum(arg))

def AT(arg, dx):
  """ AT(arg, dx):
      return 0.5 * totalT(dx * forwardSum(arg)) """
  return 0.5 * totalT(dx * forwardSum(arg))

def ATA(arg, dx):
  """ ATA(arg, dx):
      return AT(A(arg, dx), dx) """
  return AT(A(arg, dx), dx)

def diagATA(dx):
  """ diagATA(dx):
      Returns the diagonal of the 'ATA' matrix. """
  dxSum = totalT(dx)[1:]
  return np.array([0.5 * (0.5 * dx[i+1] * (2. * dx[i] + dx[i+1]) +
                          (dx[i] + dx[i+1]) * dxSum[i])
                   for i in xrange(0, len(dx) - 1)] + [0])
