'''
logoGen.py
A script utilizing the binfull and binless algorithms to create a logo.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
from svg.path import *
from xml.dom import minidom
import numpy as np
import bisect
from matplotlib import pyplot as plt

from scipy.sparse import spdiags
from scipy.stats import distributions
from scipy.sparse.linalg import LinearOperator, lgmres, spsolve
from numpy.linalg import norm

import binless
import binfull
from utilities.sampleFunctions import randomGenerator
randomGenerator.setSeed('Hi there.')

# import the data from the logo
#logo = minidom.parse('logo/debinningLogo_firstTry.svg')
logo = minidom.parse('logo/debinningLogo.svg')
paths = {path.getAttribute('id'):path.getAttribute('d')
         for path in logo.getElementsByTagName('path')}
logoPdf = parse_path(paths['logoPdf'])
coords = [0.]
alongPath = 0.
while alongPath < 1.:
  alongPath += 0.0000001
  point = logoPdf.point(alongPath)
  if point.real > coords[-1].real + 0.000001:
    coords.append(point)
while coords[0].real <= 0.000001:
  coords.pop(0)

# convert the data into the standard which debinning is used to
x = np.array([point.real for point in coords])
y = [point.imag for point in coords]
y = max(y) - np.array(y)

###### PLOT THE PDF AS A SANITY CHECK ######
fig, axes = plt.subplots(figsize=(8.5, 5.4))
fig.subplots_adjust(left=0.01, right=0.99, top=0.99, bottom=0.01)
plt.plot(x, y, '-g', linewidth=3)
plt.axis('off')
#plt.savefig('logo/functionLogo_firstTry.pdf', bbox_inches='tight')
plt.savefig('logo/functionLogo.pdf', bbox_inches='tight')
plt.close()


###### FAKE DATA SET UP ######
# Fake data looks funny... Let's do this by hand.
dx, foo, bar = binless.makeDerivativeArrays(x)
cdf = binless.A(y, dx)

# bisect_left(cdf, p) locates the left-most entry in cdf which is >= p
pdf = np.array([x[bisect.bisect_left(cdf, p)]
                for p in (randomGenerator.rand(10000) * cdf[-1])])
pdf = binless.uniqueSort(pdf)

###### PLOT THE FAKE DATA HISTOGRAM AS A SANITY CHECK ######
fig, axes = plt.subplots(figsize=(8.5, 5.4))
fig.subplots_adjust(left=0.01, right=0.99, top=0.99, bottom=0.01)
plt.hist(pdf, bins=100, color='#66a61e')
plt.axis('off')
#plt.savefig('logo/histogramLogo_firstTry.pdf', bbox_inches='tight')
plt.savefig('logo/histogramLogo.pdf', bbox_inches='tight')
plt.close()


###### BINLESS SETUP ######
# cdf of the data:
cdf = binless.dataCdf(pdf)
# derivative operators
dx, forwardDiff, delForward = binless.makeDerivativeArrays(pdf)
# diagonal of 'ATA', involving the integration operators
diagATA = binless.diagATA(dx)
# binless parameters
param = {'logAlpha': 3.5, 'beta': 1e-00,
         'eps1': 1e-12, 'eps2': 1e+01, 'end': 50}
# initial guess
u = distributions.norm.pdf(pdf, 70, 40) * len(pdf)
delu = forwardDiff * u
absu = spdiags((delu**2 + param['beta'])**(-0.5), [0], len(pdf), len(pdf))
# L operator
L = -delForward * absu * forwardDiff
# gradient
def g(LL, uu):
  return (binless.AT(binless.A(uu, dx) - cdf, dx)
          + (10. ** param['logAlpha']) * (LL * uu))
# hessian
def H(uu):
  return binless.ATA(uu, dx) + ((10. ** param['logAlpha']) * (L * uu))
# g acting on initial guess, u
gu = g(L, u)
iteration = 0
s = 1000
# binless loop
while (iteration < param['end'] and norm(s) > param['eps1']
       and norm(gu) > param['eps2']):
  iteration += 1
  # compute the step, s, towards the solution of the binless algorithm
  P = spdiags(diagATA, [0], len(pdf), len(pdf)) + (10. ** param['logAlpha']) * L
  def Pinv(xin):
    return spsolve(P, xin)
  MM = LinearOperator((len(pdf), len(pdf)), matvec=Pinv, dtype=float)
  HH = LinearOperator((len(pdf), len(pdf)), matvec=H, dtype=float)
  s, info = lgmres(HH, -gu, M=MM, tol=1e-03, maxiter=10, inner_m=30,
                      outer_k=3)
  error = norm(H(s) + gu)/(norm(gu) + 1)
  # adjust the guess with the s
  u = u + s
  # update the operators
  delu = forwardDiff * u
  absu = spdiags((delu**2 + param['beta'])**(-0.5), [0], len(pdf), len(pdf))
  L = -delForward * absu * forwardDiff
  gu = g(L, u)
  # iteration info printing
  if iteration == 1 :
    print 'iter  error     |step|   |gu|   info'
  print '%4d %8.1e %8.1e %8.1e %4d' % (iteration, error, norm(s), norm(gu), info)

###### BINLESS PLOTTING ######
fig, axes = plt.subplots(figsize=(8.5, 5.4))
fig.subplots_adjust(left=0.01, right=0.99, top=0.99, bottom=0.01)
plt.plot(pdf, u, '-b', linewidth=3)
plt.axis('off')
#plt.savefig('logo/binlessLogo_firstTry.pdf', bbox_inches='tight')
plt.savefig('logo/binlessLogo.pdf', bbox_inches='tight')
plt.close()


###### BINFULL SETUP ######
binfullWidth = (max(x) - min(x)) / 500.
binfullRange = np.arange(min(x), max(x) + 0.5*binfullWidth, binfullWidth)
binfullSmoother = binfull.linearExpandingGaussianSmoother(50000.)
binfullPdf = binfull.binfullHistogram(binfullRange, pdf, 10000000, binfullSmoother)

###### BINFULL PLOTTING ######
fig, axes = plt.subplots(figsize=(8.5, 5.4))
fig.subplots_adjust(left=0.01, right=0.99, top=0.99, bottom=0.01)
binfullPdf.fillPlot(color='#a41666')
plt.axis('off')
#plt.savefig('logo/binfullLogo_firstTry.pdf', bbox_inches='tight')
plt.savefig('logo/binfullLogo.pdf', bbox_inches='tight')
plt.close()

