'''
runBinless.py
A script testing the binless algorithm on some sample functions.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
###### IMPORTS ######
from time import time
totalTimer = time()
import utilities.sampleFunctions as funcs
import utilities.minimizers as minis
import binless

import numpy as np
import bisect
from matplotlib import pyplot as plt
plt.rc('mathtext', fontset='stix')

from scipy.sparse import spdiags
from scipy.stats import distributions
from scipy.sparse.linalg import LinearOperator, lgmres, spsolve
from numpy.linalg import norm

import sys, getopt, os

setup = {'xlow':0, 'xhigh':200, 'binWidth':5, 'step':0.0001, 'npoints':1000,
         'func':'all', 'dataDir':'dataSet_a', 'seed':'seed_a'}

###### COMMAND LINE ARGS AND SETUP ######
try:
  opts, args = getopt.getopt(sys.argv[1:], '',
                             [key + '=' for key in setup])
except getopt.GetoptError:
  print 'runBinless.py may be run with the following options:'
  print '  --xlow=[a], --xhigh=[b]: Set the x-axis range to [a,b]'
  print '  --binWidth=[#]: Set the width of the bins for the histogram'
  print '  --step=[#]: Set the step-size for fake data generation'
  print '  --npoints=[#]: Set the number of points for fake data generation'
  print '  --func=[#/all]: Choose the function by int from 0 to 4'
  print '  --dataDir=[None/directory]: Data to or from ./savedData/directory'
  print '  --seed=[string]: Set the random seed (uses hash value)'
  print '### DEFAULTS:'
  for key, value in setup.iteritems():
    print '  ' + key + '=' + str(value)
  print
  sys.exit(2)
for opt, arg in opts:
  if arg.isdigit():
    setup[opt.strip('-')] = int(arg)
  elif arg.replace('.', '', 1).isdigit():
    setup[opt.strip('-')] = float(arg)
  else:
    setup[opt.strip('-')] = arg

# saving, loading, and plotting settings, creating directories as necessary.
if setup['dataDir']:
  dataDir = 'savedData/' + setup['dataDir'] + '/'
  plotDir = 'resultPlots/' + setup['dataDir'] + '/'
  setupFileName = dataDir.partition(setup['dataDir'])
  setupFileName = setupFileName[0] + 'setup_' + setupFileName[1] + '.txt'

  # create the data directory, if necessary
  if not os.path.isdir(dataDir):
    os.mkdir(dataDir)

  # load the settings, if the settings file exists
  if os.path.exists(setupFileName):
    setupFile = open(setupFileName, 'r')
    line = (setupFile.readline().strip('\n')).partition(' = ')
    while not line == ('', '', ''):
      if line[2].isdigit():
        setup[line[0]] = int(line[2])
      elif line[2].replace('.', '', 1).isdigit():
        setup[line[0]] = float(line[2])
      else:
        setup[line[0]] = line[2]
      line = (setupFile.readline().strip('\n')).partition(' = ')
    setupFile.close()
else:
  plotDir = 'resultPlots/' + S['seed'] + 'N' + str(S['npoints']) + '/'

# make the plotting directory, if necessary
if not os.path.isdir(plotDir):
  os.mkdir(plotDir)

# shorthand for the setup dict
S = setup
# list of functions to test
if S['func'] == 'all':
  funcList = range(len(funcs.allFunctions)) 
else:
  funcList = [S['func']]
# names of the data arrays to be saved or loaded
dataNames = ['x', 'y', 'cdf', 'pdf', 'bgX', 'bgY', 'bgCdf', 'bgPdf']



###### BIG FUNC LOOP ######
for f in funcList:
  # convert func int choice to actual function
  func = funcs.allFunctions[f]

  # filename prefix for this function
  prefix = funcs.nameMap[func]

  # init random number generator with S['seed'] and the function string.
  funcs.randomGenerator.setSeed(S['seed'] + prefix)

  ###### FAKE DATA SETUP ######
  t = time()
  if S['dataDir'] and all(
        [os.path.exists(dataDir + prefix + '_' + name + '.npy')
            for name in dataNames]):
    # fake data exists; load it!
    print '\n ~~~~~~ Now loading function data: '
    print prefix
    print ' with', S['npoints'], 'points ~~~~~~ \n'
    for name in dataNames:
      exec(name + ' = np.load(dataDir + prefix + "_' + name + '.npy")')
  else:
    # generate fake data!
    print '\n ###### Now generating data for function: '
    print prefix
    print ' with ', S['npoints'], 'points ###### \n'
    x, y, cdf, pdf = funcs.functionToData(func, S['npoints'], S['step'])
    bgNpoints = [S['npoints'], cdf[-1]]
    bgX, bgY, bgCdf, bgPdf = funcs.functionToData(funcs.bgMap[func],
                                                  bgNpoints, S['step'])

  setLog = True if funcs.bgMap[func] == funcs.exponentialBG else False
  histogramRange = np.arange(S['xlow'], 0.5*S['binWidth'] + S['xhigh'],
                             S['binWidth'])
  yNormed = y / cdf[-1]
  bgYNormed = bgY / cdf[-1]
  bgYSelfNormed = bgY / bgCdf[-1]

  # Remove duplicates from pdfs -- needed for Binless algorithm
  pdf = binless.uniqueSort(pdf)
  bgPdf = binless.uniqueSort(bgPdf)
  npoints = len(pdf)
  bgNpoints = len(bgPdf)


###### PLOTTING ######
  fig, axes = plt.subplots(figsize=(11,7))
  fig.subplots_adjust(left=0.14, right=0.95, top=0.95, bottom=0.125)
  # save the plot painters as pXxxx
  # the histogram painter is a list of "patches". grab any patch with [2][0]
  pHist = (plt.hist(pdf, bins=histogramRange, log=setLog,
                    normed=True, alpha=0.4, color='#66a61e',
                    label=r'$%i$ Points' % S['npoints'], zorder=0))[2][0]
  pBG, = plt.plot(bgX, bgYNormed, '--', color='#964a01', linewidth=4,
                  label='BG only PDF', zorder=1)
  pTrue, = plt.plot(x, yNormed, '#964a01', linewidth=4,
                    label='True PDF', zorder=2)

  xwidth = S['xhigh'] - S['xlow']
  xdiv = xwidth / 4.
  axes.set_xlim(S['xlow'] - 0.002*xwidth, S['xhigh'] + 0.001*xwidth)
  axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv))
  axes.set_xticklabels([r'$%i$' % int(num) for num in
                        np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv)],
                       fontsize=30)
  axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv/5.),
                  minor=True)

  if setLog:
    axes.set_ylim(bottom=0.98e-4, top=0.102)
    axes.set_yticklabels([r'$10^{%i}$' % int(num) for num in
                          np.arange(-5,1,1)], fontsize=30)
  else:
    # This is a bit of a pain in the ass... but makes a pretty yaxis:
    ymaxx = 1.6 * plt.axis()[3]
    ydiv = 10**(np.log10(ymaxx) // 1)
    if ydiv * 3. > ymaxx:
      ydiv /= 2.
    elif ydiv * 5. <= ymaxx:
      ydiv *= 2.
    axes.set_ylim(bottom=-0.0001, top=ymaxx)
    axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv))
    axes.set_yticklabels([r'$%0.3f$' % num for num
                          in np.arange(0, ymaxx*1.01, ydiv)], fontsize=30)
    axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv/5.), minor=True)

  axes.set_xlabel('Arbitrary Units', labelpad=5, fontsize=28)
  axes.set_ylabel('Probability', labelpad=5, fontsize=28)

  axes.tick_params(length=10, width=2, labelsize=28, zorder=10, pad=10)
  axes.tick_params(which='minor', length=5, width=2, zorder=11)
  axes.minorticks_on()

  plt.legend((pTrue, pBG, pHist),
             (pTrue.get_label(), pBG.get_label(), pHist.get_label()),
             shadow=True, loc='best', prop={'size':26},
             labelspacing=0.25, handlelength=1.2, handletextpad=0.35)

  filename = plotDir + prefix + '_data.pdf'
  plt.savefig(filename)
  plt.close()
  print 'Time after fake data setup: ' + str(time() - t)


###### FAKE DATA SAVING TO FILES ######
  if S['dataDir']:
    t = time()
    for name in dataNames:
      if not os.path.exists(dataDir + prefix + '_' + name + '.npy'):
        eval('np.save(dataDir + prefix + "_' + name + '", ' + name + ')')
    print 'Time after fake data saving: ' + str(time() - t)


###### BINLESS GENERATION ######
  t = time()
  # filename prefix for this smoother and function
  binlessPrefix = prefix + '_binless'

  if all([S['dataDir'], binlessPrefix + 'LogAlpha' in setup,
          os.path.exists(dataDir + binlessPrefix + '.npy')]):
    # binless data exists; load it!
    print '\n ~~~~~~ Now loading binless data from: '
    print binlessPrefix
    print ' ~~~~~~ \n'
    binlessPdf = np.load(dataDir + binlessPrefix + '.npy')
    logAlpha = setup[binlessPrefix + 'LogAlpha']
  else:
    # left-most entry in bgX which is >= bgPdf[i]
    bgXindices = [bisect.bisect_left(bgX, value) for value in bgPdf]

    # compute cumulative distribution function from the data
    cdf = binless.dataCdf(pdf)
    bgCdf = binless.dataCdf(bgPdf)
    
    #################################################
    # Set up Numerical Calculus Operators from data #
    #################################################
    dx, forwardDiff, delForward = binless.makeDerivativeArrays(pdf)
    bgDx, bgForwardDiff, bgDelForward = binless.makeDerivativeArrays(bgPdf)

    diagATA = binless.diagATA(dx)
    bgDiagATA = binless.diagATA(bgDx)
    print 'Time after Binless loop setup: ' + str(time() - t)

    ###### BINLESS PARAMETER FINDING ######
    # Set up the iteration parameters
    '''
    logAlpha : controls the noise reduction (smoothing)
    beta  : prevents the divergence of delu / norm(delu)
    eps1  : controls the stopping criterion of pathetically small steps in the iteration
    eps2  : controls the stopping criterion of pathetically small gradient g
    end   : controls the stopping criterion of too many iterations
    '''
    param = {'logAlpha': [1., 12.], 'beta': 1e-00,
             'eps1': 1e-12, 'eps2': 1e+01, 'end': 50}

    # Use a least squares fit to find the best smoothing parameter
    def minimizeMe(logAlpha):
      ti = time()
      #####################
      # Binless Algorithm #
      #####################

      # Set the initial guess
      u = distributions.norm.pdf(bgPdf, 70, 40) * bgNpoints
      # Compute initial gradient g
      delu = bgForwardDiff * u
      absu = spdiags((delu**2 + param['beta'])**(-0.5), [0], bgNpoints, bgNpoints)
      L = -bgDelForward * absu * bgForwardDiff

      def g(LL, uu):
        return (binless.AT(binless.A(uu, bgDx) - bgCdf, bgDx)
                + (10. ** logAlpha) * (LL * uu))
      def H(uu):
        return binless.ATA(uu, bgDx) + ((10. ** logAlpha) * (L * uu))
      # g acting on initial guess, u
      gu = g(L, u)
      print 'Time after algorithm setup: ' + str(time() - ti)
      iteration = 0
      s = 1000
      while (iteration < param['end'] and norm(s) > param['eps1']
             and norm(gu) > param['eps2']):
        iteration += 1
        # compute the step, s, towards the solution of the binless algorithm
        P = spdiags(bgDiagATA, [0], bgNpoints, bgNpoints) + (10. ** logAlpha) * L
        def Pinv(xin):
          return spsolve(P, xin)
        MM = LinearOperator((bgNpoints, bgNpoints), matvec=Pinv, dtype=float)
        HH = LinearOperator((bgNpoints, bgNpoints), matvec=H, dtype=float)
        s, info = lgmres(HH, -gu, M=MM, tol=1e-03, maxiter=10, inner_m=30,
                            outer_k=3)
        error = norm(H(s) + gu)/(norm(gu) + 1)
        # adjust the guess with the s
        u = u + s
        # update the operators
        delu = bgForwardDiff * u
        absu = spdiags((delu**2 + param['beta'])**(-0.5), [0], bgNpoints, bgNpoints)
        L = -bgDelForward * absu * bgForwardDiff
        gu = g(L, u)
        # iteration info printing
        if iteration == 1 :
          print 'iter  time   error     |step|   |gu|   info'
        print '%4d %5.1f %8.1e %8.1e %8.1e %4d' % (iteration, time() - ti, error,
                                                   norm(s), norm(gu), info)

      # Binless algorithm complete! Print convergence status
      if norm(gu) > param['eps2'] :
        if norm(s) <= param['eps1'] :
          print 'Failed to converge... Step size too small.'
        elif iteration == param['end'] :
          print 'Failed to converge... Too many iterations.'
      else :
        print 'CONVERGED.'
      # in any case, compute and return chisq
      u = (1. / float((binless.A(u, bgDx))[-1])) * u
      chisq = 0
      for i, value in enumerate(u):
        chisq += (bgYSelfNormed[bgXindices[i]] - value) ** 2

      print '****************'
      print 'This evaluation:'
      print '  smoothing logAlpha:', logAlpha
      print '  absolute chisq:', chisq
      print 'Time after mimimizeMe evaluation: ' + str(time() - ti)
      return chisq

    print '\n ###### Now running binless algorithm with:'
    print binlessPrefix
    print ' ###### \n'

    logAlpha = minis.goldenSectionSearch(minimizeMe, param['logAlpha'], 0.001)
    print
    print 'Minimization complete!! Best smoothing logAlpha:', logAlpha
    print

    # Using the found best smoothing parameter
    #####################
    # Binless Algorithm #
    #####################

    # Set the initial guess
    u = distributions.norm.pdf(pdf, 70, 40) * npoints
    # Compute initial gradient g
    delu = forwardDiff * u
    absu = spdiags((delu**2 + param['beta'])**(-0.5), [0], npoints, npoints)
    L = -delForward * absu * forwardDiff

    def g(LL, uu):
      return (binless.AT(binless.A(uu, dx) - cdf, dx)
              + (10. ** logAlpha) * (LL * uu))
    def H(uu):
      return binless.ATA(uu, dx) + ((10. ** logAlpha) * (L * uu))
    # g acting on initial guess, u
    gu = g(L, u)
    print 'Time after algorithm setup: ' + str(time() - t)
    iteration = 0
    s = 1000
    while (iteration < param['end'] and norm(s) > param['eps1']
           and norm(gu) > param['eps2']):
      iteration += 1
      # compute the step, s, towards the solution of the binless algorithm
      P = spdiags(diagATA, [0], npoints, npoints) + (10. ** logAlpha) * L
      def Pinv(xin):
        return spsolve(P, xin)
      MM = LinearOperator((npoints, npoints), matvec=Pinv, dtype=float)
      HH = LinearOperator((npoints, npoints), matvec=H, dtype=float)
      s, info = lgmres(HH, -gu, M=MM, tol=1e-03, maxiter=10, inner_m=30,
                          outer_k=3)
      error = norm(H(s) + gu)/(norm(gu) + 1)
      # adjust the guess with the s
      u = u + s
      # update the operators
      delu = forwardDiff * u
      absu = spdiags((delu**2 + param['beta'])**(-0.5), [0], npoints, npoints)
      L = -delForward * absu * forwardDiff
      gu = g(L, u)
      # iteration info printing
      if iteration == 1 :
        print 'iter  time   error     |step|   |gu|   info'
      print '%4d %5.1f %8.1e %8.1e %8.1e %4d' % (iteration, time() - t, error,
                                                 norm(s), norm(gu), info)

    # Binless algorithm complete! Print convergence status
    if norm(gu) > param['eps2'] :
      if norm(s) <= param['eps1'] :
        print 'Failed to converge... Step size too small.'
      elif iteration == param['end'] :
        print 'Failed to converge... Too many iterations.'
    else :
      print 'CONVERGED.'
    binlessPdf = (1. / float((binless.A(u, dx))[-1])) * u
  print 'Time after binless setup: ' + str(time() - t)
  print

  t = time()
###### PLOTTING ######
  fig, axes = plt.subplots(figsize=(11,7))
  fig.subplots_adjust(left=0.14, right=0.95, top=0.95, bottom=0.125)
  # save the plot painters as pXxxx, or use proxy painters:
  # the histogram painter is a list of "patches". grab any patch with [2][0]
  pHist = (plt.hist(pdf, bins=histogramRange, log=setLog,
                    normed=True, alpha=0.4, color='#66a61e',
                    label=r'$%i$ Points' % S['npoints'], zorder=0))[2][0]
  pTrue, = plt.plot(x, yNormed, '#d95f02', linewidth=4,
                    label='True PDF', zorder=3)

  ex = int(np.log10(logAlpha) // 1)
  if ex:
    logAlpha *= 10**-ex
    logAlpha = r'$\log_{10}(\alpha) = %0.3f \times 10^{%i}$' % (logAlpha, ex)
  else:
    logAlpha = r'$\log_{10}(\alpha) = %0.3f$' % logAlpha
  proxyLogalpha, = plt.plot(x[:1], yNormed[:1], 'w', linewidth=0, zorder=1,
                            label=logAlpha)
  pBinless, = plt.plot(pdf, binlessPdf, 'o', zorder=2, alpha=0.55,
                       color='#a41666', label='Binless', ms=10)

  axes.set_xlim(S['xlow'] - 0.002*xwidth, S['xhigh'] + 0.001*xwidth)
  axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv))
  axes.set_xticklabels([r'$%i$' % int(num) for num in
                        np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv)],
                       fontsize=30)
  axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv/5.),
                  minor=True)

  if setLog:
    axes.set_ylim(bottom=0.98e-4, top=0.102)
    axes.set_yticklabels([r'$10^{%i}$' % int(num) for num in
                          np.arange(-5,1,1)], fontsize=30)
  else:
    axes.set_ylim(bottom=-0.0001, top=ymaxx)
    axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv))
    axes.set_yticklabels([r'$%0.3f$' % num for num
                          in np.arange(0, ymaxx*1.01, ydiv)], fontsize=30)
    axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv/5.), minor=True)

  axes.set_xlabel('Arbitrary Units', labelpad=5, fontsize=28)
  axes.set_ylabel('Probability', labelpad=5, fontsize=28)

  axes.tick_params(length=10, width=2, labelsize=28, zorder=10, pad=10)
  axes.tick_params(which='minor', length=5, width=2, zorder=11)
  axes.minorticks_on()

  plt.legend((pTrue, pHist, pBinless, proxyLogalpha),
             (pTrue.get_label(), pHist.get_label(),
              pBinless.get_label(), proxyLogalpha.get_label()),
             shadow=True, loc='best', prop={'size':26},
             labelspacing=0.25, handlelength=1.2, handletextpad=0.35)

  filename = plotDir + binlessPrefix + '.pdf'
  plt.savefig(filename)
  plt.close()

  print 'Time after plotting: ' + str(time() - t)
  print


###### BINLESS SAVING TO A FILE ######
  if S['dataDir'] and not os.path.exists(dataDir + binlessPrefix + '.npy'):
    t = time()
    np.save(dataDir + binlessPrefix, binlessPdf)
    print 'Time after binless saving: ' + str(time() - t)
    print

###### FINALIZATIONS ######
# save the settings, and the logAlpha results:
if S['dataDir']:
  setupFile = open(setupFileName, 'w')
  for key, value in setup.iteritems():
    setupFile.write(key + ' = ' + str(value) + '\n')
  setupFile.close()

print 'Time after everything: ' + str(time() - totalTimer)
