'''
minimizers.py
Two minimization routines.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
import math

###### MINIMIZATION ROUTINES ######
def crappyOneDMinimizer(func, guess, step, xtol=0.001):
  lastFunc = func(guess)
  lastGuess = guess
  guess += step
  thisFunc = func(guess)
  deltaF = (lastFunc - thisFunc) / abs(lastFunc)
  lastDeltaF = 0.
  while abs(step) > xtol:
    if deltaF < 0:
      step *= -0.1
    elif lastDeltaF < deltaF:
      step *= 2.
    lastFunc = thisFunc
    guess += step
    thisFunc = func(guess)
    lastDeltaF = deltaF
    deltaF = (lastFunc - thisFunc) / abs(lastFunc)
  return guess

phi = (1 + math.sqrt(5)) / 2
resphi = 2 - phi

def goldenSectionSearch(func, guess, tau, evaluated=None):
  """ goldenSectionSearch(func, guess, tau):
    Returns the golden section search result of a unimodal function.
    'guess' should be a list of two or three values, which correspond to
        a, (b,) and c below.
    (the middle point 'b' will be calculated, if it is not given)
    The range [guess[0], guess[2]] should contain the minimum of func.
    'tau' controls the stopping criterion:
        abs(guess[2] - guess[0]) < tau * (abs(guess[1]) + abs(x)),
        where 'x' is the next step of the algorithm

    [From http://en.wikipedia.org/wiki/Golden_section_search:
      // a and c are the current bounds; the minimum is between them.
      // b is a center point
      // f(x) is some mathematical function elsewhere defined
      // a corresponds to x1; b corresponds to x2; c corresponds to x3
      // x corresponds to x4
      // tau is a tolerance parameter; see above
    ]
  """
  # Initialize the missing value of guess, if necessary
  if len(guess) == 2:
    guess.insert(-1, (guess[1] + phi * guess[0]) / (phi + 1))
  if len(guess) != 3:
    raise ValueError("guess must be a list of floats with length 3!")

  if evaluated == None:
    evaluated = [func(value) for value in guess]
    print str(guess), str(evaluated)

  # Determine the next point in the algorithm, x
  if guess[2] - guess[1] > guess[1] - guess[0]:
    x = guess[1] + resphi * (guess[2] - guess[1])
  else:
    x = guess[1] - resphi * (guess[1] - guess[0])

  # Check for convergence
  if abs(guess[2] - guess[0]) < tau * (abs(guess[1]) + abs(x)):
    return (guess[2] + guess[0]) / 2

  # Set up the guess for the next iteration
  fx = func(x)
  assert fx != evaluated[1], "Is this function really unimodal??"
  if fx < evaluated[1]:
    if guess[2] - guess[1] > guess[1] - guess[0]:
      guess = [guess[1], x, guess[2]]
      evaluated = [evaluated[1], fx, evaluated[2]]
    else:
      guess = [guess[0], x, guess[1]]
      evaluated = [evaluated[0], fx, evaluated[1]]
  else:
    if guess[2] - guess[1] > guess[1] - guess[0]:
      guess = [guess[0], guess[1], x]
      evaluated = [evaluated[0], evaluated[1], fx]
    else:
      guess = [x, guess[1], guess[2]]
      evaluated = [fx, evaluated[1], evaluated[2]]
  return goldenSectionSearch(func, guess, tau, evaluated)
